from ats_base.log.logger import logger

from ats_base.service import build_in
from ats_case.case import base, asb
from ats_case.case.context import Context


class ATS(object):
    def __init__(self):
        self._name = 'ats'
        self._operation = None
        self._parameter = None
        self._glo = None
        self._stx = None
        self._ptx = None
        self._jp = None
        self._secs = 0
        self._result = None

    def operation(self, command: str):
        self._operation = command
        return self

    def parameter(self, param=None):
        self._parameter = param
        return self

    def glo(self, g=None):
        self._glo = g
        return self

    def stx(self, c=None):
        self._stx = c
        return self

    def ptx(self, c=None):
        self._ptx = c
        return self

    def jp(self, j=None):
        self._jp = j
        return self

    def secs(self, s=0):
        self._secs = s
        return self

    def build_in(self, context: Context):
        if isinstance(self._operation, str) and len(self._operation) > 0:
            self._parameter = base.transform(context, self._parameter)
            logger.info('~ @ATS:BUILD-IN-> operation:{} parameter:{}'.format(self._operation, self._parameter))
            try:
                self._result = build_in.handle(function=self._operation, data=self._parameter,
                                               debug_url=context.debug_service_url.get('build_in'))
                result = base.build_in_result(self._operation, self._parameter, self._result, 1)
                base.send(context, todo={'app:show': {'msg': result}})
            except Exception as e:
                result = base.build_in_result(self._operation, self._parameter, self._result, 2, str(e))
                base.send(context, todo={'app:show': {'msg': result}})

            context.runtime.sas[context.runtime.step] = result

    def flush(self, context: Context):
        if isinstance(self._glo, dict) and len(self._glo) > 0:
            logger.info('~ @ATS:GLOBAL-> global:{} result:{}'.format(self._glo, self._result))
            for result, name in self._glo.items():
                context.runtime.glo[name] = self._result[result]

    def set(self, context: Context):
        if isinstance(self._stx, dict) and len(self._stx) > 0:
            logger.info('~ @ATS:SET-> context:{} result:{}'.format(self._stx, self._result))
            for result, sx in self._stx.items():
                exec('{} = {}'.format(sx, self._result[result]))

    def put(self, context: Context):
        if isinstance(self._ptx, dict) and len(self._ptx) > 0:
            logger.info('~ @ATS:PUT-> context:{} result:{}'.format(self._ptx, self._result))
            for px, result in self._ptx.items():
                exec('{} = {}'.format(px, result))

    def jump(self, context: Context):
        if isinstance(self._jp, dict) and len(self._jp) > 0:
            step = self._jp.get('step')
            times = self._jp.get('times')

            times = base.transform(context, times)

            logger.info('~ @ATS:JUMP-> context:{} step:{} times:{}'.format(self._stx, step, times))

            context.runtime.jump_times += 1

            result = base.jump_result(context.runtime.jump_times, times, step)
            base.send(context, todo={'app:show': {'msg': result}})
            context.runtime.sas[context.runtime.step] = result

            if context.runtime.jump_times >= times:
                context.runtime.step_jump = False
                context.runtime.jump_times = 0
            else:
                context.runtime.step_jump = True
                context.runtime.step = step
                context.runtime.loop_sn = self._loop_sn(context, step)
                context.runtime.loop_index = 0

    def _loop_sn(self, context: Context, step: int):
        loops = context.case.control.get('loops')
        if loops is None or type(loops) is not list or len(loops) <= 0:
            return 0

        index = 0
        for loop in loops:
            ranges = loop.get('range')
            s = ranges.split(':')
            start_step = int(s[0])
            if step <= start_step:
                break
            index += 1

        return index

    def sleep(self, context: Context):
        self._secs = base.transform(context, self._secs)
        if isinstance(self._secs, int) and self._secs > 0:
            base.sleep(self._secs)

    def exec(self, context: Context):
        self.build_in(context)
        self.flush(context)
        self.set(context)
        self.jump(context)
        self.sleep(context)



from ats_base.common import func
from ats_base.log.logger import logger

from ats_case.case import command
from ats_case.case.context import Context


def monitoring_control(func):
    """
    逻辑控制语句监测和控制
    :param func:
    :return:
    """

    def wrap(self, index):
        # 执行前
        self._context.runtime.step = self._steps[index]

        is_exec = True
        if _is_control_start_step(self._context):
            if not _condition_is_true(self._context):
                # 条件False -> 跳到控制语句结束步骤
                is_exec = False

        # 执行步骤
        if is_exec:
            func(self, index)

        # 执行后
        not_incr_index = False  # 加1 - 步骤数组的索引
        if _is_control_end_step(self._context):
            not_incr_index = True  # 不加1 - 步骤数组的索引
            if _is_if_end_step(self._context):
                if _is_loop_end_step(self._context):
                    if _loop_is_over(self._context, self._steps):
                        # 循环没有结束, 函数体内会跳步; 结束, 函数体内不跳步, 下面 当前步骤索引加1取下一步骤号, 看是否溢出
                        not_incr_index = _is_not_overflow(self._context, self._steps)  # 没有溢出, 函数体内会跳步
                else:
                    not_incr_index = _is_not_overflow(self._context, self._steps)  # 没有溢出, 函数体内会跳步
            else:
                if _is_loop_end_step(self._context):
                    if _loop_is_over(self._context, self._steps):  # 循环没有结束, 函数体内会跳步
                        not_incr_index = _is_not_overflow(self._context, self._steps)  # 没有溢出, 函数体内会跳步

        if not_incr_index:
            index = self._steps.index(self._context.runtime.step)
        else:
            index = self._steps.index(self._context.runtime.step) + 1

        return index

    return wrap


def _is_control_start_step(context: Context):
    """
    是否是控制语句开始步骤
    :param context:
    :return:
    """
    return context.runtime.step in context.runtime.start_conditions


def _condition_is_true(context: Context):
    """
    条件判断
    :param context:
    :return:
    """
    # if条件
    ifBody = context.runtime.logicBody.get("IF_S_{}".format(context.runtime.step))
    if ifBody is not None and isinstance(ifBody, dict):
        try:
            ci = int(ifBody.get("condition"))
        except:
            ci = int(context.runtime.glo.get(ifBody.get("condition"), 1))

        if ci == 0:  # 条件判断False
            start = ifBody.get("start")
            end = ifBody.get("end")

            context.runtime.step = end
            for i in range(start, end + 1):
                context.runtime.sos.update({i: func.to_dict(result=None)})

            return False

    # loop条件 - 实例化currentLoop
    loopBody = context.runtime.logicBody.get("LOOP_S_{}".format(context.runtime.step))
    if loopBody is not None and isinstance(loopBody, dict):
        start = loopBody.get("start", 0)
        end = loopBody.get("end", 0)
        count = loopBody.get("count", 0)

        if count <= 0:
            context.runtime.step = end
            for i in range(start, end + 1):
                context.runtime.sos.update({i: func.to_dict(result=None)})
            return False

        if context.runtime.currentLoop is None:
            context.runtime.currentLoop = _new_loop(context, start=start, end=end, count=count)
            _log(context)
        else:
            # 新的循环
            if context.runtime.currentLoop.start != start:
                newLoop = _new_loop(context, start=start, end=end, count=count)

                # 判断新的循环是否在当前循环内
                if context.runtime.currentLoop.start < newLoop.start \
                        and newLoop.end <= context.runtime.currentLoop.end:
                    newLoop.parent = context.runtime.currentLoop
                else:  # 判断新循环是否与当前循环在同一个父循环内
                    if context.runtime.currentLoop.parent is not None:
                        if context.runtime.currentLoop.parent.start < newLoop.start \
                                and newLoop.end <= context.runtime.currentLoop.parent.end:
                            newLoop.parent = context.runtime.currentLoop.parent

                context.runtime.currentLoop = newLoop
                _log(context)

    return True


def _is_control_end_step(context: Context):
    """
    是否是控制语句结束步骤
    :param context:
    :return:
    """
    return context.runtime.step in context.runtime.end_conditions


def _is_if_end_step(context: Context):
    """
    是否是if的结束步骤
    :param context:
    :return:
    """
    ifBody = context.runtime.logicBody.get("IF_E_{}".format(context.runtime.step))
    if ifBody is not None and isinstance(ifBody, dict):
        return True

    return False


def _is_loop_end_step(context: Context):
    """
    是否是loop的结束步骤
    :param context:
    :return:
    """
    loopBody = context.runtime.logicBody.get("LOOP_E_{}".format(context.runtime.step))
    if loopBody is not None and isinstance(loopBody, dict):
        return True

    return False


def _loop_is_over(context: Context, steps, is_over=True):
    """
    loop是否结束
    :param context:
    :param steps:
    :param is_over:
    :return:
    """
    if context.runtime.currentLoop is not None:
        context.runtime.currentLoop.index += 1
        context.runtime.loop_index = context.runtime.currentLoop.index  # 兼容2.0.0以前的版本

        if context.runtime.currentLoop.index < context.runtime.currentLoop.count:  # 当前循环未结束, 跳到第一步
            context.runtime.step = context.runtime.currentLoop.start
            is_over = False

            _log(context, level=2)  # 再循环 - 输出日志
        else:
            _log(context, level=1)  # 循环结束 - 输出日志

            if context.runtime.currentLoop.parent is not None:
                context.runtime.currentLoop = context.runtime.currentLoop.parent

                context.runtime.loop_sn = context.runtime.currentLoop.sn
                context.runtime.loop_count = context.runtime.currentLoop.count
                context.runtime.loop_index = context.runtime.currentLoop.index

                if context.runtime.step >= context.runtime.currentLoop.end:
                    is_over = _loop_is_over(context, steps)
                else:
                    context.runtime.step = steps[steps.index(context.runtime.step) + 1]
                    is_over = False

        if is_over:
            context.runtime.currentLoop = None

            context.runtime.loop_sn = 0
            context.runtime.loop_count = 0
            context.runtime.loop_index = 0

    return is_over


def _is_not_overflow(context: Context, steps):
    """
    是否溢出
    :param context:
    :param steps:
    :return:
    """
    try:
        context.runtime.step = steps[steps.index(context.runtime.step) + 1]
    except:  # 溢出测试用例步骤
        return False

    return True


def _new_loop(context: Context, start=0, end=0, count=0):
    context.runtime.incr_loop_sn += 1
    context.runtime.loop_sn = context.runtime.incr_loop_sn
    # 兼容2.0.0以前的版本
    if context.runtime.currentLoop is None and context.runtime.loop_count > 0:
        count = context.runtime.loop_count
    return CurrentLoop(sn=context.runtime.incr_loop_sn, start=start, end=end, count=count)


def _log(context: Context, level=0):
    if level == 0:
        logger.info('~ @TCC-LOOP-> loops[#{}] start. -range {}:{}  -count {}'.format(
            context.runtime.currentLoop.sn, context.runtime.currentLoop.start,
            context.runtime.currentLoop.end, context.runtime.currentLoop.count))

        command.client().message('[#{}]循环开始 - 步骤范围[{}-{}], 共{}次'.format(
            context.runtime.currentLoop.sn, context.runtime.currentLoop.start,
            context.runtime.currentLoop.end, context.runtime.currentLoop.count)).show(context)

        logger.info('~ @TCC-LOOP-> loops[#{}-{}:{}], -count {}, -index {}'.format(
            context.runtime.currentLoop.sn, context.runtime.currentLoop.start,
            context.runtime.currentLoop.end, context.runtime.currentLoop.count,
            context.runtime.currentLoop.index + 1))
        command.client().message('[#{}-{}:{}]循环 - 共{}次, 当前执行第{}次'.format(
            context.runtime.currentLoop.sn, context.runtime.currentLoop.start,
            context.runtime.currentLoop.end, context.runtime.currentLoop.count,
            context.runtime.currentLoop.index + 1)).show(context)

    if level == 1:
        command.client().message("[#{}-{}:{}]循环结束...".format(context.runtime.currentLoop.sn,
                                                             context.runtime.currentLoop.start,
                                                             context.runtime.currentLoop.end)).show(context)
        logger.info('~ @TCC-LOOP-> loops[#{}-{}:{}] end.'.format(context.runtime.currentLoop.sn,
                                                                 context.runtime.currentLoop.start,
                                                                 context.runtime.currentLoop.end))

    if level == 2:
        logger.info('~ @TCC-LOOP-> loops[#{}-{}:{}], -count {}, -index {}'.format(
            context.runtime.currentLoop.sn, context.runtime.currentLoop.start,
            context.runtime.currentLoop.end, context.runtime.currentLoop.count,
            context.runtime.currentLoop.index + 1))
        command.client().message('[#{}-{}:{}]循环 - 共{}次, 当前执行第{}次'.format(
            context.runtime.currentLoop.sn, context.runtime.currentLoop.start,
            context.runtime.currentLoop.end, context.runtime.currentLoop.count,
            context.runtime.currentLoop.index + 1)).show(context)


class CurrentLoop(object):
    def __init__(self, sn=0, start=0, end=0, count=0):
        self._sn = sn
        self._start = start
        self._end = end
        self._count = count
        self._index = 0
        self._parent = None

    @property
    def sn(self):
        return self._sn

    @sn.setter
    def sn(self, value):
        self._sn = value

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @property
    def count(self):
        return self._count

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, value):
        self._index = value

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value



from ats_base.common import func
from ats_base.common.enum import ProClazz
from ats_base.log.logger import logger
from ats_base.service import em, udm

from ats_case.case import base, asb
from ats_case.case.context import Context


class Encryptor(object):
    def __init__(self, protocol):
        self._protocol = ProClazz(protocol)
        self._operation = None
        self._parameter = None
        self._result = None

    def operation(self, op: str):
        self._operation = op
        return self

    def parameter(self, param=None):
        self._parameter = param
        return self

    def handle(self, context: Context):
        try:
            if self._parameter is None:
                self._parameter = {}
            self._parameter = context.runtime.sos[context.runtime.step - 1]['result']
            self._parameter['session_key'] = context.test_sn
        except:
            pass

        logger.info(
            '~ @EM-> security:{} operation:{} parameter:{}'.format(self._protocol, self._operation, self._parameter))
        self._result = em.handle(self._protocol.name, self._operation, self._parameter)
        logger.info('~ @EM<- security:{} operation:{} result:{}'.format(self._protocol, self._operation, self._result))
        self._flush(context)

    def _flush(self, context: Context):
        context.runtime.sos.update({context.runtime.step: func.to_dict(obj='em', op=self._operation
                                                                       , parameter=self._parameter,
                                                                       result=self._result)})

    def acv(self, context: Context):
        data = func.to_dict(result=self._result)

        logger.info('~ @ACD-> module:{} function:{} parameter:{}'.format(
            self._protocol.name, self._operation, self._parameter))
        result = udm.handle(module='em.{}'.format(self._protocol.name), function=self._operation,
                            data=data, debug_url=context.debug_service_url.get('acd'))
        logger.info('~ @ACD<- module:{} function:{} result:{}'.format(self._protocol.name, self._operation, result))

        context.runtime.sas[context.runtime.step] = result

        return result

    def exec(self, context: Context):
        self.handle(context)
        base.send(context, todo={'app:show': {'msg': self.acv(context)}})



from ats_base.log.logger import logger

from ats_case.case import base, asb
from ats_case.case.context import Context


class App(object):
    def __init__(self):
        self._name = 'app'
        self._operation = None
        self._message = None
        self._parameter = None

    def operation(self, command: str):
        self._operation = command
        return self

    def message(self, msg):
        self._message = {'msg': msg}
        return self

    def parameter(self, param=None):
        self._parameter = param
        return self

    def show(self, context: Context, types=2):
        logger.info('~ @APP-> operation:{} message:{}'.format('show', self._message))
        base.send(context, todo={'{}:{}'.format(self._name, 'show'): self._message},
                  types=types, end=base.end_step(context, types))

    def error(self, context: Context, types=2):
        logger.info('~ @APP-> operation:{} message:{}'.format('error', self._message))
        base.send(context, todo={'{}:{}'.format(self._name, 'error'): self._message},
                  types=types, end=base.end_step(context, types))

    def exec(self, context: Context):
        logger.info('~ @APP-> operation:{} message:{}'.format(self._operation, self._message))
        base.send(context, todo={'{}:{}'.format(self._name, self._operation): self._message})


from ats_case.case import base

from ats_case.entity.ats import ATS
from ats_case.entity.bench import Bench
from ats_case.entity.client import App
from ats_case.entity.encrypt import Encryptor
from ats_case.entity.meter import Meter

"""
    注释篇
"""


def step_annotation(**param):
    """
    测试步骤方法注释 - 装饰器
    :param param: desc-测试步骤描述
    :return:
    """
    desc = param.get('desc')

    def decorate(callback):
        def fn(*args, **kwargs):
            base.send(args[0], todo={'app:show': {'msg': desc}})
            r = callback(*args, **kwargs)
            return r

        return fn

    return decorate


"""
    通讯协议篇
"""


def meter(protocol: str):
    return Meter(protocol)


"""
    加密机篇
"""


def encrypt(protocol: str):
    return Encryptor(protocol)


"""
    表台篇
"""


def bench():
    return Bench()


"""
    测试终端篇
"""


def client():
    return App()


"""
    平台篇
"""


def ats():
    return ATS()



import pickle
import sys
from datetime import datetime

import gevent
from importlib import import_module, reload

from ats_base.common import func
from ats_base.common.enum import Payload
from ats_base.log.logger import logger
from ats_base.service import msg

from ats_case.case import asb, base, translator, command
from ats_case.case.context import Context
from ats_case.common.enum import *
from ats_case.common.error import *
from ats_case.statement import logic


def execute(context: Context):
    if context.mode == WorkMode.FORMAL:
        FormalExecutor(context).exec()
    else:
        DebugExecutor(context).exec()


class Executor(object):
    def __init__(self, context: Context):
        self._context = context
        self._model = None
        self._steps = []

    def handle(self):
        """
        抽象方法 - 子类实现脚本脚本
        :return:
        """
        pass

    def exec(self):
        """
        执行用例所有步骤
        :return:
        """
        self.handle()

        index = self._load()  # 加载所需关键变量
        while index < len(self._steps):
            index = self.step_exec(index)

    def _load(self):
        """
        加载索引或断点索引
        :return:
        """
        # 断点信息
        if self._context.renew == 1:
            try:
                rt = pickle.loads(asb.BreakPoint.get(self._context, 'runtime'))
                if isinstance(rt, self._context.Runtime):
                    self._incr_loop_sn = rt.incr_loop_sn
                    self._context.runtime.step = rt.step

                    self._context.runtime.loop_sn = rt.loop_sn
                    self._context.runtime.loop_count = rt.loop_count
                    self._context.runtime.loop_index = rt.loop_index

                    self._context.runtime.currentLoop = rt.currentLoop
                    self._context.runtime.glo = rt.glo

                    self._context.runtime.sos = rt.sos
                    self._context.runtime.sas = rt.sas

                    self._context.runtime.meter_result_valid_fail_delay = rt.meter_result_valid_fail_delay
                    self._context.runtime.bench_result_valid_fail_delay = rt.bench_result_valid_fail_delay

                    self._context.runtime.final_result = rt.final_result

                    try:
                        return self._steps.index(self._context.runtime.step)
                    except:
                        pass
            except:
                pass

        return 0

    def _flush(self):
        """
        缓存断点数据
        :return:
        """
        try:
            asb.BreakPoint.set(self._context, 'runtime', pickle.dumps(self._context.runtime))
        except:
            pass

    @logic.monitoring_control
    def step_exec(self, index: int):
        """
        执行单个步骤
        :return:
        """
        if self._context.mode == WorkMode.FORMAL:  # 项目测试 - 协程切换
            gevent.sleep(0.05)

        logger.info('~ @TCC-STEP-> steps[#{}] execute'.format(self._context.runtime.step))

        self._flush()  # 缓存在断点续测时所需关键变量
        if self.is_exec():
            try:
                getattr(self._model, 'step_{}'.format(self._context.runtime.step))(self._context)
            except APIError as ae:
                logger.info(str(ae))
                self._err_msg_notify(str(ae))

                raise AssertionError(str(ae))
            except Exception as e:
                self._context.runtime.sas[self._context.runtime.step] = '{} - {} '.format(base.CODE_ERROR, str(e))
                if isinstance(e, BenchReadNoneError):
                    asb.Application.set(self._context.test_sn, "meter:quantity",
                                        int(asb.Application.get(self._context.test_sn, "meter:quantity")) - 1)
                command.client().message(str(e)).error(self._context)
                logger.error(str(e))
                self._err_msg_notify(str(e))

                raise AssertionError(str(e))

    def is_exec(self):
        """
        判断步骤是否执行
        :return:
        """
        if base.offbench(self._context, self._context.offbench):
            return False

        return True

    def _err_msg_notify(self, err: str):
        """
        异常消息通知
        :param err:
        :return:
        """
        msg.send('dingding', func.to_dict(payload=Payload.ERROR.value, workmode=self._context.mode.value,
                                          project=self._context.project, case_name=self._context.case.name,
                                          meter_pos=self._context.meter.pos, step=self._context.runtime.step,
                                          test_sn=self._context.test_sn,
                                          start_time=self._context.start_time.strftime('%Y.%m.%d %H:%M:%S'),
                                          end_time=datetime.now().strftime('%Y.%m.%d %H:%M:%S'),
                                          msg=err))


def extract_steps(content: list):
    n_s = []
    for s in content:
        if s.upper().find('STEP_') >= 0:
            num = func.extract_digit(s)
            n_s.append(int(num))

    return sorted(n_s)


def import_script(clazz: ScriptClazz, context: Context, manual_dir: str = 'debug'):
    # 分为两种情况: 0. 手动编写的脚本 1.自动转化的脚本
    if clazz == ScriptClazz.AUTO:
        steps = translator.translate(context)
        a_name = 'script.auto.{}.tsm_{}'.format(context.tester.username.lower(), context.meter.pos)
        try:
            model = sys.modules[a_name]
            model = reload(model)
        except:
            model = import_module(a_name)
    else:
        m_name = 'script.manual.{}.{}'.format(manual_dir, context.case.steps)
        try:
            model = sys.modules[m_name]
            model = reload(model)
        except:
            model = import_module(m_name)

        steps = extract_steps(dir(model))

    return model, steps


class FormalExecutor(Executor):
    def handle(self):
        self._model, self._steps = import_script(self._context.case.script, self._context, 'formal')


class DebugExecutor(Executor):
    def handle(self):
        self._model, self._steps = import_script(self._context.case.script, self._context)



from datetime import datetime

from ats_base.common import func
from ats_base.service import mm
from ats_case.common.enum import WorkMode, ScriptClazz


class Context(object):
    """
    测试用例上下文
    """

    def __init__(self, sn):
        self._test_sn = sn
        self._parse()

    def _parse(self):
        tl = mm.Dict.get("test:log", self._test_sn)

        self._mode = WorkMode(tl.get('mode'))
        self._project = tl.get('project')
        self._renew = tl.get('renew', 0)  # 1: 断点续测
        self._offbench = tl.get('offbench', 0)  # 1: 脱机
        self._start_time = datetime.now()
        self._tester = self.Tester(tl.get('tester'))
        self._case = self.Case(tl.get('usercase'))
        self._meter = self.Meter(self, tl.get('meter'))
        self._bench = self.Bench(tl.get('bench'))
        # 运行时
        self._runtime = self.Runtime(self)
        # Debug模式 - 测试开发人员本机调试数据比对服务使用
        self._debug_service_url = tl.get('debug_service_url', {})

    @property
    def mode(self):
        return self._mode

    @property
    def project(self):
        return self._project

    @property
    def renew(self):
        return self._renew

    @property
    def offbench(self):
        return self._offbench

    @property
    def start_time(self):
        return self._start_time

    @property
    def test_sn(self):
        return self._test_sn

    @property
    def tester(self):
        return self._tester

    @property
    def bench(self):
        return self._bench

    @property
    def case(self):
        return self._case

    @property
    def meter(self):
        return self._meter

    @property
    def runtime(self):
        return self._runtime

    @property
    def debug_service_url(self):
        return self._debug_service_url

    class Tester(object):
        def __init__(self, data: dict):
            self._ip = data.get('ip')
            self._port = data.get('port')
            self._username = data.get('username')
            self._hostname = data.get('hostname')

        @property
        def ip(self):
            return self._ip

        @property
        def port(self):
            return self._port

        @property
        def username(self):
            return self._username

        @property
        def hostname(self):
            return self._hostname

        @property
        def api(self):
            return "http://{}:{}/accept".format(self.ip, self.port)

    class Bench(object):
        def __init__(self, data: dict):
            if data is not None and len(data) > 0:
                self._manufacture = data.get('manufacture', '')
                self._model = data.get('model', '')
                self._port = data.get('port', 0)
                self._error_threshold = data.get('error_threshold', 0)

        #         self._iabc = data.get('iabc', 'H')
        #
        @property
        def manufacture(self):
            return self._manufacture

        @property
        def model(self):
            return self._model

        @property
        def port(self):
            return self._port

        @property
        def error_threshold(self):
            return self._error_threshold

    class Case(object):
        def __init__(self, data: dict):
            self._id = data.get('id', -1)
            self._name = data.get('name', 'test')
            self._code = data.get('code')
            self._version = data.get('version')
            self._script = ScriptClazz(data.get('script', 1))
            self._control = data.get('control', {})
            self._steps = data.get('steps', {})
            self._end = data.get('end', 0)

            if self._code is not None:
                self._name = '{}_{}'.format(self._name, self._code)
            if self._version is not None:
                self._name = '{}_{}'.format(self._name, self._version)
            if isinstance(self._control, str):
                self._control = self._control.replace('null', 'None')
                self._control = eval(self._control)
            if isinstance(self._steps, str):
                self._steps = self._steps.replace('null', 'None')
                self._steps = eval(self._steps)

        @property
        def id(self):
            return self._id

        @property
        def name(self):
            return self._name

        @property
        def script(self):
            return self._script

        @property
        def control(self):
            return self._control

        @property
        def steps(self):
            return self._steps

        @property
        def code(self):
            return self._code

        @property
        def version(self):
            return self._version

        @property
        def end(self):
            return self._end

    class Meter(object):
        def __init__(self, parent, data: dict):
            self._parent = parent
            self._index = data.get('index', 0)
            self._pos = data.get('pos')
            self._addr = data.get('addr')
            self._no = data.get('no')
            self._protocol = data.get('protocol')
            self._model = data.get('model')
            self._connect = data.get('connect')
            self._rated_voltage = data.get('rated_voltage')
            self._rated_current = data.get('rated_current')
            self._max_current = data.get('max_current')
            self._min_current = data.get('min_current')
            self._transfer_current = data.get('transfer_current')  # 转折电流
            self._starting_current = data.get('starting_current')
            self._frequency = data.get('frequency')
            self._mconst = data.get('mconst')  # 脉冲常数
            self._mpluse = data.get('mpluse')  # 圈数
            self._iabc = data.get('iabc')
            self._end = data.get('end', 0)
            self._operator = data.get('opcode', '00000000')
            self._pwd = data.get('password', '000000')

            session = dict()
            if self._operator is not None:
                session['operator'] = self._operator
            if self._pwd is not None:
                session['pwd'] = self._pwd

            mm.Dict.put(self._parent.test_sn, 'session', session)

        #         self._channel = {'RS485': data.get('baudrate_485'), 'IR': data.get('baudrate_ir'),
        #                          'PLC': data.get('baudrate_plc')}
        #
        #     def get_channel(self, c: CHANNEL):
        #         return func.to_dict(type=c.value, baudrate=self._channel.get(c.value))

        @property
        def index(self):
            return self._index

        @property
        def pos(self):
            return self._pos

        @property
        def addr(self):
            return self._addr

        @property
        def no(self):
            return self._no

        @property
        def protocol(self):
            return self._protocol

        @property
        def model(self):
            return self._model

        @property
        def connect(self):
            return self._connect

        @property
        def rated_voltage(self):
            return self._rated_voltage

        @property
        def rated_current(self):
            return self._rated_current

        @property
        def max_current(self):
            return self._max_current

        @property
        def min_current(self):
            return self._min_current

        @property
        def transfer_current(self):
            return self._transfer_current

        @property
        def starting_current(self):
            return self._starting_current

        @property
        def frequency(self):
            return self._frequency

        @property
        def mconst(self):
            return self._mconst

        @property
        def mpluse(self):
            return self._mpluse

        @property
        def iabc(self):
            return self._iabc

        @property
        def end(self):
            return self._end

    class Runtime(object):
        """
        运行时
        """

        def __init__(self, parent):
            self._parent = parent
            # 仅供2.0.0之前版本使用
            self._loop_sn = 0
            self._loop_count = 0
            self._loop_index = 0
            # 当前步骤
            self._incr_loop_sn = 0
            self._step = -1
            # 控制语句数据
            self._start_conditions = set()
            self._end_conditions = set()
            self._logicBody = {}
            # 当前循环
            self._currentLoop = None
            # 结果验证失败, 重复操作延时
            self._meter_result_valid_fail_delay = 2
            self._bench_result_valid_fail_delay = 2
            # 步骤协议解析返回结果集
            self._sos = {}
            # 步骤对比返回结果集 - 用例结束时统计执行结果
            self._sas = {}
            # 全局变量 - 用户自定义
            self._glo = {}
            # 多帧时使用
            self._final_result = None

            # 加载逻辑控制数据
            self._logic_control()

            # 加载全局变量
            g = self._parent.case.control.get('global')
            if isinstance(g, dict) and len(g) > 0:
                self._glo = g

        def _logic_control(self):
            # if集合
            ifs = self._parent.case.control.get('ifs')
            if isinstance(ifs, dict):
                try:
                    ifs.pop('offbench')
                except:
                    pass

                if len(ifs) > 0:
                    for ranges, condition in ifs.items():
                        step_start = int(ranges.split(':')[0])
                        step_end = int(ranges.split(':')[1])
                        self._start_conditions.add(step_start)
                        self._end_conditions.add(step_end)

                        self._logicBody["IF_S_{}".format(step_start)] = func.to_dict(start=step_start, end=step_end,
                                                                                     condition=condition)
                        self._logicBody["IF_E_{}".format(step_end)] = func.to_dict(start=step_start, end=step_end,
                                                                                   condition=condition)

            # loop集合
            loops = self._parent.case.control.get('loops')
            if isinstance(loops, list):
                for loop in loops:
                    ranges = loop.get('range')
                    count = loop.get('count')

                    step_start = int(ranges.split(':')[0])
                    step_end = int(ranges.split(':')[1])
                    self._start_conditions.add(step_start)
                    self._end_conditions.add(step_end)

                    self._logicBody["LOOP_S_{}".format(step_start)] = func.to_dict(start=step_start, end=step_end,
                                                                                   count=count)
                    self._logicBody["LOOP_E_{}".format(step_end)] = func.to_dict(start=step_start, end=step_end,
                                                                                 count=count)

        @property
        def loop_sn(self):
            return self._loop_sn

        @loop_sn.setter
        def loop_sn(self, value):
            self._loop_sn = value

        @property
        def loop_count(self):
            return self._loop_count

        @loop_count.setter
        def loop_count(self, value):
            self._loop_count = value

        @property
        def loop_index(self):
            return self._loop_index

        @loop_index.setter
        def loop_index(self, value):
            self._loop_index = value

        @property
        def incr_loop_sn(self):
            return self._incr_loop_sn

        @incr_loop_sn.setter
        def incr_loop_sn(self, value):
            self._incr_loop_sn = value

        @property
        def step(self):
            return self._step

        @step.setter
        def step(self, value):
            self._step = value

        @property
        def start_conditions(self):
            return self._start_conditions

        @property
        def end_conditions(self):
            return self._end_conditions

        @property
        def logicBody(self):
            return self._logicBody

        @property
        def currentLoop(self):
            return self._currentLoop

        @currentLoop.setter
        def currentLoop(self, value):
            self._currentLoop = value

        @property
        def meter_result_valid_fail_delay(self):
            return self._meter_result_valid_fail_delay

        @meter_result_valid_fail_delay.setter
        def meter_result_valid_fail_delay(self, value):
            self._meter_result_valid_fail_delay = value

        @property
        def bench_result_valid_fail_delay(self):
            return self._bench_result_valid_fail_delay

        @bench_result_valid_fail_delay.setter
        def bench_result_valid_fail_delay(self, value):
            self._bench_result_valid_fail_delay = value

        @property
        def sos(self):
            return self._sos

        @sos.setter
        def sos(self, value):
            self._sos = value

        @property
        def sas(self):
            return self._sas

        @sas.setter
        def sas(self, value):
            self._sas = value

        @property
        def glo(self):
            return self._glo

        @glo.setter
        def glo(self, value):
            self._glo = value

        @property
        def final_result(self):
            return self._final_result

        @final_result.setter
        def final_result(self, value):
            self._final_result = value



import psutil

from multiprocessing import Process

from ats_base.log.logger import logger

from ats_case.case import asb
from ats_case.common.enum import WorkMode
from ats_case.manage.start import run


def start(data):
    """
    启动测试进程
    :param data:
    :return:
    """
    # 判断测试终端是否可以连接
    p = Process(target=run, kwargs=data)
    p.start()

    logger.info('TEST PROCESS[{}] START - DATA: {}...'.format(p.pid, data))
    return p.pid


def suspend(data):
    """
    暂停测试进程
    :param data:
    :return:
    """
    pid = data.get('pid')
    logger.info('TEST PROCESS[{}] SUSPEND...'.format(pid))

    p = psutil.Process(pid)
    p.suspend()


def resume(data):
    """
    恢复暂停测试进程
    :param data:
    :return:
    """
    pid = data.get('pid')
    logger.info('TEST PROCESS[{}] RESUME...'.format(pid))

    p = psutil.Process(pid)
    p.resume()


def cancel(data):
    """
    结束测试进程
    :param data:
    :return:
    """
    pid = data.get('pid')
    logger.info('TEST PROCESS[{}] CANCEL...'.format(pid))

    p = psutil.Process(pid)
    p.kill()


def callback(key: str, data):
    """
    回调
    :param key:
    :param data:
    :return:
    """
    mode = WorkMode[data.get('mode', 'DEBUG').upper()]
    test_sn = data.get('test_sn')
    meter_pos = data.get('meter_pos')
    if mode == WorkMode.FORMAL:
        test_sn += ':{}'.format(meter_pos)
    frame = data.get('frame')

    asb.Session.set_callback(test_sn, key, frame)





class APIError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "APIError - API服务接口[{}], 连接失败.".format(repr(self.value))


class MeterNeedSecurityAuthentication(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "MeterNeedSecurityAuthentication - 电表需要安全认证, 表位:[{}].".format(repr(self.value))


class MeterSecurityAuthenticationError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "MeterSecurityAuthenticationError - 电表安全认证失败, {}.".format(repr(self.value))


class MeterResponseFrameNoneError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "MeterResponseFrameNoneError - 电表无响应帧, {}.".format(repr(self.value))


class MeterResponseFrameParseError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "MeterResponseFrameParseError - 电表响应帧解析失败, {}.".format(repr(self.value))


class BenchReadNoneError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "BenchReadNoneError - 表台读取误差错误, {}.".format(repr(self.value))


class ClientError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "ClientReturnError - 客户端返回结果为空. {}".format(repr(self.value))

# MeterTimeoutError 电表连接超时
# 服务接口错误  电表错误  客户端错误 加密机错误
# 连接错误 代码错误 操作错误



import os
import threading

import gevent
import pytest

from datetime import datetime

from ats_base.common import func
from ats_base.common.enum import Payload
from ats_base.service import mm, db, msg

from ats_case.case import asb
from ats_case.common.enum import WorkMode


# 主进程
def run(**kwargs):
    try:
        mode = WorkMode(kwargs.get('mode'))
        if mode == WorkMode.FORMAL:
            pt = FormalMode(kwargs)
        else:
            pt = DebugMode(kwargs)
        pt.run()
    except:
        pass


# 装饰器
def message(fc):
    def wrapper(self):
        fc(self)

        # after
        try:
            self._end_time = datetime.now()
            msg.send('dingding', func.to_dict(payload=Payload.NORMAL.value, workmode=self._workmode.value,
                                              project=self._project, test_sn=self._sn,
                                              start_time=self._start_time.strftime('%Y.%m.%d %H:%M:%S'),
                                              end_time=self._end_time.strftime('%Y.%m.%d %H:%M:%S')))
        except Exception as e:
            print(str(e))

    return wrapper


class ExecMode(object):
    def __init__(self, data: dict):
        self._data = data
        self._workmode = WorkMode(data.get('mode', 'DEBUG'))
        self._project = self._data.get('project', '')
        self._username = self._data.get('tester', {}).get('username', '')
        self._sn = self._username.upper() + datetime.now().strftime('%y%m%d%H%M%S%f')
        self._start_time = datetime.now()
        self._end_time = None

        self._init()

    def _init(self):
        pass

    def run(self):
        pass

    def _build(self, work_mode: WorkMode, code: str = None):
        if code is None:
            code = 'case'

        user_dir = func.makeDir(func.project_dir(), 'testcase', work_mode.value.lower(), self._username)
        template_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'template', 'testcase_v1.tmp')
        script_file = os.path.join(user_dir, 'test_{}.py'.format(code))

        with open(template_file, 'r', encoding='UTF-8') as file:
            content = file.read()
            content = content.replace('{script}', code.upper())
        with open(script_file, 'w', encoding='UTF-8') as file:
            file.write(content)

        return script_file


class FormalMode(ExecMode):
    def _init(self):
        sn = self._data.get('test_sn')

        if sn is None or len(sn) == 0:
            self._save()
        else:  # 断点续测
            self._sn = sn
            self._data['renew'] = 1
            self._data = mm.Dict.get('test:log', self._sn)

        self._cases = self._data.get('usercases', [])
        self.meters = self._data.get('meters', [])

        asb.Application.set(self._sn, "meter:quantity", len(self.meters))

    def _save(self):
        try:
            db.save('test:log', sn=self._sn, start_time=self._start_time.strftime('%Y-%m-%d %H:%M:%S'),
                    offbench=self._data.get('offbench', 0), tester=self._data.get('tester'),
                    usercases=self._data.get('usercases'), meters=self._data.get('meters'),
                    bench=self._data.get('bench'))
        except Exception as e:
            print(str(e))

    @message
    def run(self):
        index = 0
        for id_v, data in self._cases.items():
            case = self._get_case(id_v, data)

            index += 1
            if index == len(self._cases):
                case['end'] = 1

            case_task = self.CaseTask(self, case)
            case_task.start()
            case_task.join()

    def _get_case(self, id_v, data):
        idv = id_v.split(':')
        case_id = idv[0]
        version = idv[1]

        c_v = db.query('view:case:version', id=case_id, version=version)
        if isinstance(data, dict):
            g = data.get('global')
            if isinstance(g, dict):
                if isinstance(c_v['control'], str):
                    c_v['control'] = c_v['control'].replace('null', 'None')
                    control = eval(c_v['control'])

                    control['global'] = g
                    c_v['control'] = str(control)

        return c_v

    class CaseTask(threading.Thread):
        def __init__(self, parent, case):
            super(FormalMode.CaseTask, self).__init__()
            self._parent = parent
            self._case = case

        def run(self):
            gs = []
            length = len(self._parent.meters)
            for i in range(length):
                # i = 0 执行操作表台 传入参数index
                self._parent.meters[i]['index'] = i
                # 全部用例和电表结束标签
                if self._case.get('end', 0) == 1 and i == (length - 1):
                    self._parent.meters[i]['end'] = 1
                gs.append(gevent.spawn(self._parent.exec, self._parent.handle(case=self._case,
                                                                              meter=self._parent.meters[i])))

            gevent.joinall(gs)

    def handle(self, case, meter):
        self._data['usercase'] = case
        self._data['meter'] = meter

        test_sn = '{}:{}'.format(self._sn, meter['pos'])
        mm.Dict.put('test:log', test_sn, self._data)

        return test_sn

    def exec(self, test_sn):
        pytest.main(["-sv", self._build(WorkMode.FORMAL), '--sn={}'.format(test_sn)])


class DebugMode(ExecMode):
    def _init(self):
        asb.Application.set(self._sn, "meter:quantity", 1)
        self._flush()

    def _flush(self):
        mm.Dict.put('test:log', self._sn, self._data)

    @message
    def run(self):
        pytest.main(["-sv", self._build(WorkMode.DEBUG), '--sn={}'.format(self._sn)])



import gevent

from ats_base.common import func
from ats_base.log.logger import logger
from ats_base.service import udm

from ats_case.case import base, asb
from ats_case.case.context import Context
from ats_case.common.enum import *


class Bench(object):
    def __init__(self):
        self._operation = None
        self._parameter = None
        self._is_meter_pos_operation = False  # 针对台体的单个表位, 不是针对整个台体的操作
        self._interval = None
        self._secs = 0
        self._result = None
        self._exec_times = 0
        self._func = None
        self._expect_result = None
        self._func_parameter = {}
        self._not_pass = None
        self._break_case = 0
        self._repeat_times = 0
        self._repeat_sleep = 0

    def operation(self, command: str):
        self._operation = command
        if self._operation.find('read') >= 0:
            self._is_meter_pos_operation = True
        return self

    def parameter(self, param=None):
        self._parameter = param
        if type(self._parameter) is dict:
            if self._parameter.get('meter_pos') is not None:
                self._is_meter_pos_operation = True
        return self

    def interval(self, times):
        self._interval = times
        return self

    def secs(self, s=0):
        self._secs = s
        return self

    def compare(self, data):
        self._func = data.get('code')
        self._expect_result = data.get('expect_result', None)
        self._func_parameter = data.get('parameter', {})
        self._not_pass = data.get('not_pass', None)

        if isinstance(self._not_pass, dict) and len(self._not_pass) > 0:
            self._break_case = self._not_pass.get('break_case', 0)
            self._repeat_times = self._not_pass.get('repeat_times', 0)
            self._repeat_sleep = self._not_pass.get('sleep', 0)

        return self

    def encode(self, context: Context):
        logger.info(
            '~ @BENCH-> manufacture:{} operation:{} parameter:{}'.format(context.bench.manufacture, self._operation,
                                                                         self._parameter))

        if type(self._parameter) is dict:
            self._parameter = base.transform(context, self._parameter)

    def decode(self, context: Context):
        logger.info('~ @BENCH<- manufacture:{} operation:{} result:{}'.format(context.bench.manufacture,
                                                                              self._operation, self._result))
        self._flush(context)

    def acv(self, context: Context):
        result = str(self._result)
        data = func.to_dict(result=self._result)
        if self._func is not None:
            data['parameter'] = base.transform(context, self._func_parameter)
            try:
                if self._expect_result is None:
                    data['expect_result'] = context.runtime.sos[context.runtime.step - 1]
                else:
                    data['expect_result'] = context.runtime.sos[int(self._expect_result)]
            except:
                pass

            logger.info('~ @ACD-> module:{} function:{} parameter:{}'.format('bench', self._func, data['parameter']))
            result = udm.handle(module='bench', function=self._func, data=data,
                                debug_url=context.debug_service_url.get('acd'))
            logger.info('~ @ACD<- module:{} function:{} result:{}'.format('bench', self._func, result))

            context.runtime.sas[context.runtime.step] = result

        return result

    def _flush(self, context: Context):
        context.runtime.sos.update({context.runtime.step: func.to_dict(obj='bench', op=self._operation
                                                                       , parameter=self._parameter,
                                                                       result=self._result)})

    def sleep(self, context: Context):
        self._secs = base.transform(context, self._secs)
        if isinstance(self._secs, int) and self._secs > 0:
            base.send(context, todo={'app:show': {'msg': base.BENCH_SLEEP_MESSAGE.format(self._secs)}})
            base.sleep(self._secs)

    def _times(self, context: Context):
        self._interval = base.transform(context, self._interval)
        if self._interval > 0:
            if context.runtime.loop_index == 0 or (context.runtime.loop_index + 1) % self._interval != 0:
                return False

            self._exec_times += 1
        return True

    def exec(self, context: Context):
        if context.meter.index == 0 or self._is_meter_pos_operation:
            if self._times(context):
                if not self._exec_main(context):
                    self._fail_repeat(context)

                if self._is_meter_pos_operation:
                    pass_quantity = asb.Application.get(context.test_sn, 'pass_quantity')
                    if pass_quantity == 'NULL':
                        pass_quantity = 0
                    pass_quantity = int(pass_quantity) + 1
                    asb.Application.set(context.test_sn, 'pass_quantity', pass_quantity)

                    if pass_quantity >= int(asb.Application.get(context.test_sn, "meter:quantity")):
                        asb.Application.set(context.test_sn, 'pass_quantity', 0)
                        self.sleep(context)
                else:
                    self.sleep(context)

    def _exec_main(self, context: Context):
        self.encode(context)
        try:
            self._result = base.send(context, todo={'bench:{}'.format(self._operation): self._parameter})
            self._result = self._result.get('result')
        except:
            self._result = ''
        self.decode(context)

        self._result = self.acv(context)
        base.send(context, todo={'app:show': {'msg': self._result}})

        return base.result_judge(self._result)

    def _fail_repeat(self, context: Context):
        if isinstance(self._repeat_times, int) and self._repeat_times > 0:
            is_finished = False     # 单只表比对结果是否合格
            asb.Application.set(context.test_sn, 'finished_quantity', 0)

            for r in range(self._repeat_times):
                if context.mode == WorkMode.FORMAL:  # 项目测试 - 协程切换
                    gevent.sleep(0.05)

                # 多表测试时, 多表比对结果合格数量
                finished_quantity = int(asb.Application.get(context.test_sn, 'finished_quantity'))
                # 多表测试时, 多表比对结果执行一轮不合格数量
                one_loop_finished_quantity = asb.Application.get(context.test_sn, 'one_loop_finished_quantity')
                if one_loop_finished_quantity == 'NULL':
                    one_loop_finished_quantity = 0
                else:
                    one_loop_finished_quantity = int(one_loop_finished_quantity)

                if not is_finished:
                    if self._exec_main(context):
                        is_finished = True
                        finished_quantity += 1
                        asb.Application.set(context.test_sn, 'finished_quantity', finished_quantity)
                    else:
                        one_loop_finished_quantity += 1
                        asb.Application.set(context.test_sn, 'one_loop_finished_quantity', one_loop_finished_quantity)

                # 多表测试时, 判断是否是最后一只表, 是否睡眠
                quantity = int(asb.Application.get(context.test_sn, "meter:quantity"))
                if one_loop_finished_quantity >= quantity - finished_quantity:
                    # 一轮循环最后一只表
                    asb.Application.set(context.test_sn, 'one_loop_finished_quantity', 0)
                    if finished_quantity < quantity:
                        # 最后一轮最后一只表不睡眠, 其他轮次最后一只表都睡眠
                        if isinstance(self._repeat_sleep, int) and self._repeat_sleep > 0:
                            base.sleep(self._repeat_sleep)
                        else:
                            base.sleep(context.runtime.bench_result_valid_fail_delay)

                if is_finished:
                    if finished_quantity >= quantity:
                        # 全部合格完成
                        break

            if self._break_case == 1 and not is_finished:
                raise AssertionError(str(self._result))



from ats_base.common import func

from ats_base.service import pro, em
from ats_case.case import base, asb
from ats_case.case.context import Context
from ats_case.common.error import MeterSecurityAuthenticationError


def establish(context: Context, protocol: str):
    exec('_{}(context, protocol)'.format(protocol))


def _gw698(context: Context, protocol: str, retry_times=3):
    try:
        # 1. 读取esam参数
        result = pro.encode(
            func.to_dict(protocol=protocol, comm_addr=context.meter.addr,
                         operation='get:request:normal_list',
                         element=['F1000200', 'F1000300', 'F1000400', 'F1000500', 'F1000600', 'F1000700'],
                         session_key=context.test_sn))
        frame = base.write(context, result.get('frame'))
        data = pro.decode(func.to_dict(protocol=protocol, frame=frame, session_key=context.test_sn))
        # 2. 协商密钥
        e_r = data.get('result')
        e_r['session_key'] = context.test_sn
        em.handle(protocol, 'negotiate', e_r)
        # 3. 连接
        result = pro.encode(
            func.to_dict(protocol=protocol, comm_addr=context.meter.addr,
                         operation='connect:request:symmetry',
                         element='',
                         parameter={
                             '期望的应用层协议版本号': 21,
                             '期望的协议一致性块': 'FFFFFFFFC0000000',
                             '期望的功能一致性块': 'FFFEC400000000000000000000000000',
                             '客户机发送帧最大尺寸(字节)': 512,
                             '客户机接收帧最大尺寸(字节)': 512,
                             '客户机接收帧最大窗口尺寸(个)': 1,
                             '客户机最大可处理APDU尺寸': 2000,
                             '期望的应用连接超时时间(秒)': 7200,
                             '认证请求对象': {
                                 '对称加密': {
                                     '密文1': asb.Session.get(context, 'em_data'),
                                     '客户机签名1': asb.Session.get(context, 'em_mac')
                                 }
                             }
                         },
                         session_key=context.test_sn))
        frame = base.write(context, result.get('frame'))
        data = pro.decode(func.to_dict(protocol=protocol, frame=frame, session_key=context.test_sn))
        # 4. 加密
        e_r = data.get('result')
        e_r['session_key'] = context.test_sn
        em.handle(protocol, 'secret', e_r)
        base.sleep(1)
    except Exception as e:
        retry_times -= 1
        if retry_times <= 0:
            raise MeterSecurityAuthenticationError('[表位:{}]{}'.format(context.meter.pos, str(e)))
        else:
            base.sleep(5)
            _gw698(context, protocol, retry_times)


def _dlms(context: Context, retry_times=3):
    pass


def _dlt645(context: Context, retry_times=3):
    pass


def _cjt188(context: Context, retry_times=3):
    pass



import os

from pathlib import Path

from ats_base.common import func
from ats_case.case.context import Context
from ats_case.common.enum import OperationClazz


def translate(context: Context):
    return Translator(context).write()


class Translator(object):
    def __init__(self, context: Context):
        self.KEYWORD = ["if", "for", "while"]
        self.TAB = "   "
        self.LINE = "\n"
        self._context = context

    def write(self):
        steps = []

        with open(self._script(), "w", encoding='utf-8') as f:
            f.write(self._gen_import())
            f.write(self._gen_line(2))

            for step, operations in self._context.case.steps.items():
                steps.append(int(step))
                f.write(self._gen_step(int(step), operations))

                tab_count = 1
                # for op in operations:
                f.write(self._gen_tab(tab_count) + self._gen_operation(operations) + self._gen_line(1))
                # if self._contain_keyword(step):
                #     tab_count += 1
                f.write(self._gen_line(2))

            return steps

    def _script(self):
        user_dir = func.makeDir(func.project_dir(), 'script', 'auto', self._context.tester.username)
        script_file = os.path.join(user_dir, 'tsm_{}.py'.format(self._context.meter.pos))

        if not os.path.exists(user_dir):
            Path(user_dir).mkdir(parents=True, exist_ok=True)

        return script_file

    def _gen_import(self):
        return 'from ats_case.case import command' + self.LINE + 'from ats_case.case.context import Context' + self.LINE

    def _gen_step(self, step: int, op: dict):
        return '@command.step_annotation(desc="{}"){}'.format(op.get('desc', ''), self.LINE) + \
               'def step_{}(context: Context):{}'.format(step, self.LINE)

    def _gen_tab(self, count: int):
        ts = ''
        for i in range(count):
            ts += self.TAB

        return ts

    def _gen_line(self, count: int):
        ls = ''
        for i in range(count):
            ls += self.LINE

        return ls

    def _contain_keyword(self, step: str):
        for k in self.KEYWORD:
            if k in step:
                return True

        return False

    def _gen_operation(self, op: dict):
        opt = OperationClazz(op.get('type').upper())

        code = eval('{}(self._context).translate(op)'.format(opt.name))

        return code + self.LINE


class Operation(object):
    def __init__(self, context: Context):
        self._context = context

    def translate(self, op: dict):
        pass


class METER(Operation):
    def translate(self, op: dict):
        clazz = op.get('clazz')
        module = op.get('module')
        opt = op.get('operation')
        elem = op.get('element')
        param = op.get('parameter')
        addi = op.get('addition')
        se = op.get('security')
        rp = op.get('report')
        acd = op.get('acd')
        secs = op.get('sleep')
        ci = op.get('data_clazz_id')

        addr = self._context.meter.addr
        if isinstance(module, int):
            addr = '{}:{}'.format(module, self._context.meter.addr)

        code = "command.meter('{}').comm_addr('{}').operation('{}')".format(clazz, addr, opt)

        code += _transfer('element', elem)
        code += _transfer('parameter', param)
        code += _transfer('addition', addi)
        code += _transfer('security', se)
        code += _transfer('report', rp)
        code += _transfer('compare', acd)
        code += _transfer('secs', secs)
        code += _transfer('chip_id', ci)

        code += ".exec(context)"

        return code


class EM(Operation):
    def translate(self, op: dict):
        clazz = op.get('clazz')
        opt = op.get('operation')
        param = op.get('parameter')

        code = "command.encrypt('{}').operation('{}').parameter({}).exec(context)".format(clazz, opt, param)

        return code


class BENCH(Operation):
    def translate(self, op: dict):
        opt = op.get('operation')
        param = op.get('parameter')
        interval = op.get('interval')
        acd = op.get('acd')
        secs = op.get('sleep')

        code = "command.bench().operation('{}')".format(opt)

        code += _transfer('parameter', param)
        code += _transfer('interval', interval)
        code += _transfer('compare', acd)
        code += _transfer('secs', secs)

        code += ".exec(context)"

        return code


class APP(Operation):
    def translate(self, op: dict):
        clazz = op.get('clazz')
        opt = op.get('operation')
        msg = op.get('message')
        param = op.get('parameter')

        code = "command.app('{}').operation('{}')".format(clazz, opt)

        if msg is not None:
            if param is not None:
                msg = msg.format(param)
            code += ".message('{}')".format(msg)

        code += ".exec(context)"

        return code


class ATS(Operation):
    def translate(self, op: dict):
        opt = op.get('operation')
        param = op.get('parameter')
        glo = op.get('cache')
        stx = op.get('set')
        ptx = op.get('put')
        jp = op.get('jump')
        secs = op.get('sleep')

        code = "command.ats().operation('{}')".format(opt)

        code += _transfer('parameter', param)
        code += _transfer('glo', glo)
        code += _transfer('stx', stx)
        code += _transfer('ptx', ptx)
        code += _transfer('jp', jp)
        code += _transfer('secs', secs)

        code += ".exec(context)"

        return code


def _transfer(attribute, value):
    if value is not None:
        if isinstance(value, str):
            code = ".{}('{}')".format(attribute, value)
        else:
            code = ".{}({})".format(attribute, value)

        return code

    return ''



import re
import gevent

from ats_base.common import func
from ats_base.common.enum import ProClazz
from ats_base.log.logger import logger
from ats_base.service import pro, udm

from ats_case.case import base, asb
from ats_case.case.context import Context
from ats_case.common.enum import *
from ats_case.common.error import *
from ats_case.security import link


class Meter(object):
    def __init__(self, protocol):
        self._protocol = ProClazz(protocol)
        self._comm_addr = None
        self._operation = None
        self._element = None
        self._parameter = None
        self._addition = None
        self._security = None
        self._report = None
        self._is_report = False
        self._report_sleep = 0
        self._secs = 0
        self._chip_id = None
        self._frame = None
        self._parse = None
        self._func_module = None
        self._func = None
        self._expect_result = None
        self._func_parameter = {}
        self._not_pass = None
        self._break_case = 0
        self._repeat_times = 0
        self._repeat_sleep = 0

    def comm_addr(self, addr: str):
        self._comm_addr = addr
        return self

    def operation(self, op: str):
        self._operation = op
        return self

    def element(self, di):
        self._element = di
        return self

    def parameter(self, param=None):
        self._parameter = param
        return self

    def addition(self, addi=None):
        self._addition = addi
        return self

    def security(self, se=None):
        self._security = se
        return self

    def report(self, rp=None):
        self._report = rp
        if isinstance(self._report, dict) and len(self._report) > 0:
            self._is_report = True
            try:
                self._report_sleep = int(self._report.get('sleep', 0))
            except:
                pass
        return self

    def secs(self, s=0):
        self._secs = s
        return self

    def chip_id(self, ci):
        self._chip_id = ci
        return self

    def compare(self, data):
        self._func_module = data.get('module')
        self._func = data.get('code')
        self._expect_result = data.get('expect_result', None)
        self._func_parameter = data.get('parameter', {})
        self._not_pass = data.get('not_pass', None)

        if isinstance(self._not_pass, dict) and len(self._not_pass) > 0:
            self._break_case = self._not_pass.get('break_case', 0)
            self._repeat_times = self._not_pass.get('repeat_times', 0)
            self._repeat_sleep = self._not_pass.get('sleep', 0)

        return self

    def frame(self, hexStr: str):
        self._frame = hexStr
        return self

    def encode(self, context: Context):
        logger.info(
            '~ @PRO-ENCODE-> security:{} comm_addr:{} operation:{} element:{}'.format(self._protocol,
                                                                                      self._comm_addr,
                                                                                      self._operation,
                                                                                      self._element))

        self._element = base.transform(context, self._element)
        self._parameter = base.transform(context, self._parameter)
        self._addition = base.transform(context, self._addition)

        parse = pro.encode(func.to_dict(protocol=self._protocol.name, comm_addr=self._comm_addr,
                                        operation=self._operation, element=self._element,
                                        chip_id=self._chip_id, parameter=self._parameter,
                                        addition=self._addition, security=self._security,
                                        session_key=context.test_sn))
        # 1. 是否需要安全认证
        if parse.get('need_sa') == 1:
            raise MeterNeedSecurityAuthentication(context.meter.pos)

        logger.info('~ @PRO-ENCODE<- security:{} frame:{}'.format(self._protocol, parse.get('frame')))

        self._frame = parse.get('frame')
        return self._frame

    def decode(self, context: Context, index=0):
        # 1. 异常判断 - 无响应帧
        if self._frame is None or len(self._frame) <= 0:
            self._parse = base.METER_NO_RESPONSE_FRAME
            self._flush(context)
            raise MeterResponseFrameNoneError('表位:{}'.format(context.meter.pos))
            # return self._parse

        logger.info('~ @PRO-DECODE-> security:{} frame:{}'.format(self._protocol, self._frame))
        try:
            data = pro.decode(
                func.to_dict(protocol=self._protocol.name, frame=self._frame, session_key=context.test_sn))
        except:
            # 2. 异常判断 - 响应帧解析错误
            self._parse = base.METER_ERROR_DECODE
            self._flush(context)
            raise MeterResponseFrameParseError('表位:{}'.format(context.meter.pos))
            # return self._parse
        logger.info('~ @PRO-DECODE<- security:{} parse:{}'.format(self._protocol, data))

        # 3. 是否需要安全认证
        if data.get('need_sa') == 1:
            raise MeterNeedSecurityAuthentication(context.meter.pos)

        # 4. 异常判断 - 操作电表失败
        if data.get('error') == 1:
            self._parse = base.METER_ERROR_OPERATION
            self._flush(context)
            return self._parse
            # raise MeterOperationError(data.get('result'))

        # 5. 分帧处理开始
        next_frame = data.get('next_frame', None)
        if next_frame is not None:
            self._frame = base.write(context, next_frame)
            self.decode(context, index + 1)
        else:
            context.runtime.final_result = data.get('result')
        # 5. 分帧处理结束

        if index == 0:
            self._parse = context.runtime.final_result
            self._flush(context)
            return self._parse

    def _flush(self, context: Context):
        context.runtime.sos.update({context.runtime.step: func.to_dict(obj='meter', op=self._operation
                                                                       , element=self._element
                                                                       , parameter=self._parameter,
                                                                       result=self._parse)})

    def sleep(self, context: Context):
        self._secs = base.transform(context, self._secs)
        if isinstance(self._secs, int) and self._secs > 0:
            base.send(context, todo={'app:show': {'msg': base.METER_SLEEP_MESSAGE.format(self._secs)}})
            base.sleep(self._secs)

    def acv(self, context: Context):
        result = str(self._parse)
        if self._func is not None:
            if type(self._func_parameter) is dict:
                self._func_parameter['mode'] = context.mode.name
                self._func_parameter = base.transform(context, self._func_parameter)
            self._expect_result_handle(context)

            data = func.to_dict(result=self._parse, expect_result=self._expect_result,
                                parameter=self._func_parameter)

            logger.info('~ @ACD-> module:{} function:{} parameter:{}'.format(
                self._func_module, self._func, self._func_parameter))
            result = udm.handle(module='device.{}'.format(self._func_module), function=self._func, data=data,
                                debug_url=context.debug_service_url.get('acd'))
            logger.info('~ @ACD<- module:{} function:{} result:{}'.format(self._func_module, self._func, result))

            context.runtime.sas[context.runtime.step] = result

        return result

    def _expect_result_handle(self, context: Context):
        try:
            if self._expect_result is None or len(str(self._expect_result).strip()) <= 0:
                self._expect_result = context.runtime.sos[context.runtime.step - 1]
            else:
                rs = re.findall(r"(\d+)", str(self._expect_result))
                if len(rs) == 1:
                    self._expect_result = context.runtime.sos[int(rs[0])]
                else:
                    self._expect_result = []
                    for r in rs:
                        self._expect_result.append(context.runtime.sos[int(r)])
        except:
            pass

    def exec(self, context: Context):
        if self._is_report:
            self._exec_report(context)
        else:
            self._exec_main(context)
        self._fail_repeat(context)
        self.sleep(context)

    def _exec_report(self, context: Context):
        base.sleep(self._report_sleep)
        try:
            self._frame = base.report(context)
        except:
            self._frame = None

        self.decode(context)
        self._result = self.acv(context)
        base.send(context, todo={'app:show': {'msg': self._result}})

    def _exec_main(self, context: Context, retry_times=3):
        try:
            self.encode(context)
            try:
                self._frame = base.write(context, self._frame)
            except:
                self._frame = None

            try:
                self.decode(context)

                self._result = self.acv(context)
                base.send(context, todo={'app:show': {'msg': self._result}})
            except MeterResponseFrameNoneError as ne:
                self._redo(context, ne, retry_times)
            except MeterResponseFrameParseError as pe:
                self._redo(context, pe, retry_times)
        except MeterNeedSecurityAuthentication as ms:
            try:
                link.establish(context, self._protocol.name.lower())
                self._exec_main(context)
            except MeterSecurityAuthenticationError as ma:
                raise ma

    def _redo(self, context: Context, e, retry_times):
        retry_times -= 1
        if retry_times <= 0:
            self._result = self.acv(context)
            base.send(context, todo={'app:show': {'msg': self._result}})
        else:
            base.send(context, todo={'app:show': {'msg': str(e)}})
            base.sleep(5)
            self._exec_main(context, retry_times)

    def _fail_repeat(self, context: Context):
        if not base.result_judge(self._result):
            if isinstance(self._repeat_times, int) and self._repeat_times > 0:
                is_finished = False  # 单只表比对结果是否合格
                for r in range(self._repeat_times):
                    if context.mode == WorkMode.FORMAL:  # 项目测试 - 协程切换
                        gevent.sleep(0.05)

                    if self._is_report:
                        self._exec_report(context)
                    else:
                        self._exec_main(context)

                    if base.result_judge(self._result):
                        is_finished = True
                        break

                    if isinstance(self._repeat_sleep, int) and self._repeat_sleep > 0:
                        base.sleep(self._repeat_sleep)
                    else:
                        base.sleep(context.runtime.meter_result_valid_fail_delay)

                if self._break_case == 1 and not is_finished:
                    raise AssertionError(str(self._result))
